from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from git import Repo

from utilities.pathlib import to_path

if TYPE_CHECKING:
    from utilities.types import MaybeCallablePathLike


def get_repo(path: MaybeCallablePathLike = Path.cwd, /) -> Repo:
    """Get the repo object."""
    return Repo(to_path(path), search_parent_directories=True)


__all__ = ["get_repo"]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from functools import cache
from logging import LogRecord
from statistics import fmean
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Self,
    SupportsFloat,
    TypedDict,
    assert_never,
    cast,
    overload,
    override,
)
from zoneinfo import ZoneInfo

from whenever import (
    Date,
    DateDelta,
    DateTimeDelta,
    PlainDateTime,
    Time,
    TimeDelta,
    Weekday,
    YearMonth,
    ZonedDateTime,
)

from utilities.dataclasses import replace_non_sentinel
from utilities.functions import get_class_name
from utilities.math import sign
from utilities.platform import get_strftime
from utilities.sentinel import Sentinel, sentinel
from utilities.tzlocal import LOCAL_TIME_ZONE, LOCAL_TIME_ZONE_NAME
from utilities.zoneinfo import UTC, to_time_zone_name

if TYPE_CHECKING:
    from utilities.types import (
        DateOrDateTimeDelta,
        DateTimeRoundMode,
        Delta,
        MaybeCallableDateLike,
        MaybeCallableTimeLike,
        MaybeCallableZonedDateTimeLike,
        TimeOrDateTimeDelta,
        TimeZoneLike,
    )


# bounds


ZONED_DATE_TIME_MIN = PlainDateTime.MIN.assume_tz(UTC.key)
ZONED_DATE_TIME_MAX = PlainDateTime.MAX.assume_tz(UTC.key)


DATE_TIME_DELTA_MIN = DateTimeDelta(
    weeks=-521722,
    days=-5,
    hours=-23,
    minutes=-59,
    seconds=-59,
    milliseconds=-999,
    microseconds=-999,
    nanoseconds=-999,
)
DATE_TIME_DELTA_MAX = DateTimeDelta(
    weeks=521722,
    days=5,
    hours=23,
    minutes=59,
    seconds=59,
    milliseconds=999,
    microseconds=999,
    nanoseconds=999,
)
DATE_DELTA_MIN = DATE_TIME_DELTA_MIN.date_part()
DATE_DELTA_MAX = DATE_TIME_DELTA_MAX.date_part()
TIME_DELTA_MIN = TimeDelta(hours=-87831216)
TIME_DELTA_MAX = TimeDelta(hours=87831216)


DATE_TIME_DELTA_PARSABLE_MIN = DateTimeDelta(
    weeks=-142857,
    hours=-23,
    minutes=-59,
    seconds=-59,
    milliseconds=-999,
    microseconds=-999,
    nanoseconds=-999,
)
DATE_TIME_DELTA_PARSABLE_MAX = DateTimeDelta(
    weeks=142857,
    hours=23,
    minutes=59,
    seconds=59,
    milliseconds=999,
    microseconds=999,
    nanoseconds=999,
)
DATE_DELTA_PARSABLE_MIN = DateDelta(days=-999999)
DATE_DELTA_PARSABLE_MAX = DateDelta(days=999999)


DATE_TWO_DIGIT_YEAR_MIN = Date(1969, 1, 1)
DATE_TWO_DIGIT_YEAR_MAX = Date(DATE_TWO_DIGIT_YEAR_MIN.year + 99, 12, 31)


## common constants


ZERO_DAYS = DateDelta()
ZERO_TIME = TimeDelta()
MICROSECOND = TimeDelta(microseconds=1)
MILLISECOND = TimeDelta(milliseconds=1)
SECOND = TimeDelta(seconds=1)
MINUTE = TimeDelta(minutes=1)
HOUR = TimeDelta(hours=1)
DAY = DateDelta(days=1)
WEEK = DateDelta(weeks=1)
MONTH = DateDelta(months=1)
YEAR = DateDelta(years=1)


##


def add_year_month(x: YearMonth, /, *, years: int = 0, months: int = 0) -> YearMonth:
    """Add to a year-month."""
    y = x.on_day(1) + DateDelta(years=years, months=months)
    return y.year_month()


def sub_year_month(x: YearMonth, /, *, years: int = 0, months: int = 0) -> YearMonth:
    """Subtract from a year-month."""
    y = x.on_day(1) - DateDelta(years=years, months=months)
    return y.year_month()


##


@dataclass(repr=False, order=True, unsafe_hash=True, kw_only=False)
class DatePeriod:
    """A period of dates."""

    start: Date
    end: Date

    def __post_init__(self) -> None:
        if self.start > self.end:
            raise DatePeriodError(start=self.start, end=self.end)

    def __add__(self, other: DateDelta, /) -> Self:
        """Offset the period."""
        return self.replace(start=self.start + other, end=self.end + other)

    def __contains__(self, other: Date, /) -> bool:
        """Check if a date/datetime lies in the period."""
        return self.start <= other <= self.end

    @override
    def __repr__(self) -> str:
        cls = get_class_name(self)
        return f"{cls}({self.start}, {self.end})"

    def __sub__(self, other: DateDelta, /) -> Self:
        """Offset the period."""
        return self.replace(start=self.start - other, end=self.end - other)

    def at(
        self, obj: Time | tuple[Time, Time], /, *, time_zone: TimeZoneLike = UTC
    ) -> ZonedDateTimePeriod:
        """Combine a date with a time to create a datetime."""
        match obj:
            case Time() as time:
                start = end = time
            case Time() as start, Time() as end:
                ...
            case never:
                assert_never(never)
        tz = to_time_zone_name(time_zone)
        return ZonedDateTimePeriod(
            self.start.at(start).assume_tz(tz), self.end.at(end).assume_tz(tz)
        )

    @property
    def delta(self) -> DateDelta:
        """The delta of the period."""
        return self.end - self.start

    def format_compact(self) -> str:
        """Format the period in a compact fashion."""
        fc, start, end = format_compact, self.start, self.end
        if self.start == self.end:
            return f"{fc(start)}="
        if self.start.year_month() == self.end.year_month():
            return f"{fc(start)}-{fc(end, fmt='%d')}"
        if self.start.year == self.end.year:
            return f"{fc(start)}-{fc(end, fmt='%m%d')}"
        return f"{fc(start)}-{fc(end)}"

    @classmethod
    def from_dict(cls, mapping: PeriodDict[Date] | PeriodDict[dt.date], /) -> Self:
        """Convert the dictionary to a period."""
        match mapping["start"]:
            case Date() as start:
                ...
            case dt.date() as py_date:
                start = Date.from_py_date(py_date)
            case never:
                assert_never(never)
        match mapping["end"]:
            case Date() as end:
                ...
            case dt.date() as py_date:
                end = Date.from_py_date(py_date)
            case never:
                assert_never(never)
        return cls(start=start, end=end)

    def replace(
        self, *, start: Date | Sentinel = sentinel, end: Date | Sentinel = sentinel
    ) -> Self:
        """Replace elements of the period."""
        return replace_non_sentinel(self, start=start, end=end)

    def to_dict(self) -> PeriodDict[Date]:
        """Convert the period to a dictionary."""
        return PeriodDict(start=self.start, end=self.end)

    def to_py_dict(self) -> PeriodDict[dt.date]:
        """Convert the period to a dictionary."""
        return PeriodDict(start=self.start.py_date(), end=self.end.py_date())


@dataclass(kw_only=True, slots=True)
class DatePeriodError(Exception):
    start: Date
    end: Date

    @override
    def __str__(self) -> str:
        return f"Invalid period; got {self.start} > {self.end}"


##


def datetime_utc(
    year: int,
    month: int,
    day: int,
    /,
    hour: int = 0,
    minute: int = 0,
    second: int = 0,
    millisecond: int = 0,
    microsecond: int = 0,
    nanosecond: int = 0,
) -> ZonedDateTime:
    """Create a UTC-zoned datetime."""
    nanos = int(1e6) * millisecond + int(1e3) * microsecond + nanosecond
    return ZonedDateTime(
        year,
        month,
        day,
        hour=hour,
        minute=minute,
        second=second,
        nanosecond=nanos,
        tz=UTC.key,
    )


##


@overload
def diff_year_month(
    x: YearMonth, y: YearMonth, /, *, years: Literal[True]
) -> tuple[int, int]: ...
@overload
def diff_year_month(
    x: YearMonth, y: YearMonth, /, *, years: Literal[False] = False
) -> int: ...
@overload
def diff_year_month(
    x: YearMonth, y: YearMonth, /, *, years: bool = False
) -> int | tuple[int, int]: ...
def diff_year_month(
    x: YearMonth, y: YearMonth, /, *, years: bool = False
) -> int | tuple[int, int]:
    """Compute the difference between two year-months."""
    x_date, y_date = x.on_day(1), y.on_day(1)
    diff = x_date - y_date
    if years:
        yrs, mth, _ = diff.in_years_months_days()
        return yrs, mth
    mth, _ = diff.in_months_days()
    return mth


##


def format_compact(
    obj: Date | Time | PlainDateTime | ZonedDateTime,
    /,
    *,
    fmt: str | None = None,
    path: bool = False,
) -> str:
    """Format the date/datetime in a compact fashion."""
    match obj:
        case Date() as date:
            obj_use = date.py_date()
            fmt_use = "%Y%m%d" if fmt is None else fmt
        case Time() as time:
            obj_use = time.round().py_time()
            fmt_use = "%H%M%S" if fmt is None else fmt
        case PlainDateTime() as date_time:
            obj_use = date_time.round().py_datetime()
            fmt_use = "%Y%m%dT%H%M%S" if fmt is None else fmt
        case ZonedDateTime() as date_time:
            plain = format_compact(date_time.to_plain(), fmt=fmt)
            tz = date_time.tz
            if path:
                tz = tz.replace("/", "~")
            return f"{plain}[{tz}]"
        case never:
            assert_never(never)
    return obj_use.strftime(get_strftime(fmt_use))


##


def from_timestamp(i: float, /, *, time_zone: TimeZoneLike = UTC) -> ZonedDateTime:
    """Get a zoned datetime from a timestamp."""
    return ZonedDateTime.from_timestamp(i, tz=to_time_zone_name(time_zone))


def from_timestamp_millis(i: int, /, *, time_zone: TimeZoneLike = UTC) -> ZonedDateTime:
    """Get a zoned datetime from a timestamp (in milliseconds)."""
    return ZonedDateTime.from_timestamp_millis(i, tz=to_time_zone_name(time_zone))


def from_timestamp_nanos(i: int, /, *, time_zone: TimeZoneLike = UTC) -> ZonedDateTime:
    """Get a zoned datetime from a timestamp (in nanoseconds)."""
    return ZonedDateTime.from_timestamp_nanos(i, tz=to_time_zone_name(time_zone))


##


def get_now(time_zone: TimeZoneLike = UTC, /) -> ZonedDateTime:
    """Get the current zoned date-time."""
    return ZonedDateTime.now(to_time_zone_name(time_zone))


NOW_UTC = get_now(UTC)


def get_now_local() -> ZonedDateTime:
    """Get the current zoned date-time in the local time-zone."""
    return get_now(LOCAL_TIME_ZONE)


NOW_LOCAL = get_now_local()


def get_now_plain(time_zone: TimeZoneLike = UTC, /) -> PlainDateTime:
    """Get the current plain date-time."""
    return get_now(time_zone).to_plain()


NOW_PLAIN = get_now_plain()


def get_now_local_plain() -> PlainDateTime:
    """Get the current plain date-time in the local time-zone."""
    return get_now_local().to_plain()


NOW_LOCAL_PLAIN = get_now_local_plain()


##


def get_time(time_zone: TimeZoneLike = UTC, /) -> Time:
    """Get the current time."""
    return get_now(time_zone).time()


TIME_UTC = get_time(UTC)


def get_time_local() -> Time:
    """Get the current time in the local time-zone."""
    return get_time(LOCAL_TIME_ZONE)


TIME_LOCAL = get_time_local()


##


def get_today(time_zone: TimeZoneLike = UTC, /) -> Date:
    """Get the current, timezone-aware local date."""
    return get_now(time_zone).date()


TODAY_UTC = get_today(UTC)


def get_today_local() -> Date:
    """Get the current, timezone-aware local date."""
    return get_today(LOCAL_TIME_ZONE)


TODAY_LOCAL = get_today_local()


##


def is_weekend(
    date_time: ZonedDateTime,
    /,
    *,
    start: tuple[Weekday, Time] = (Weekday.SATURDAY, Time.MIN),
    end: tuple[Weekday, Time] = (Weekday.SUNDAY, Time.MAX),
) -> bool:
    """Check if a datetime is in the weekend."""
    weekday, time = date_time.date().day_of_week(), date_time.time()
    start_weekday, start_time = start
    end_weekday, end_time = end
    if start_weekday.value == end_weekday.value:
        return start_time <= time <= end_time
    if start_weekday.value < end_weekday.value:
        return (
            ((weekday == start_weekday) and (time >= start_time))
            or (start_weekday.value < weekday.value < end_weekday.value)
            or ((weekday == end_weekday) and (time <= end_time))
        )
    return (
        ((weekday == start_weekday) and (time >= start_time))
        or (weekday.value > start_weekday.value)
        or (weekday.value < end_weekday.value)
        or ((weekday == end_weekday) and (time <= end_time))
    )


##


def mean_datetime(
    datetimes: Iterable[ZonedDateTime],
    /,
    *,
    weights: Iterable[SupportsFloat] | None = None,
) -> ZonedDateTime:
    """Compute the mean of a set of datetimes."""
    datetimes = list(datetimes)
    match len(datetimes):
        case 0:
            raise MeanDateTimeError from None
        case 1:
            return datetimes[0]
        case _:
            timestamps = [d.timestamp_nanos() for d in datetimes]
            timestamp = round(fmean(timestamps, weights=weights))
            return ZonedDateTime.from_timestamp_nanos(timestamp, tz=datetimes[0].tz)


@dataclass(kw_only=True, slots=True)
class MeanDateTimeError(Exception):
    @override
    def __str__(self) -> str:
        return "Mean requires at least 1 datetime"


##


def min_max_date(
    *,
    min_date: Date | None = None,
    max_date: Date | None = None,
    min_age: DateDelta | None = None,
    max_age: DateDelta | None = None,
    time_zone: TimeZoneLike = UTC,
) -> tuple[Date | None, Date | None]:
    """Compute the min/max date given a combination of dates/ages."""
    today = get_today(time_zone)
    min_parts: list[Date] = []
    if min_date is not None:
        min_parts.append(min_date)
    if max_age is not None:
        min_parts.append(today - max_age)
    min_date_use = max(min_parts, default=None)
    max_parts: list[Date] = []
    if max_date is not None:
        max_parts.append(max_date)
    if min_age is not None:
        max_parts.append(today - min_age)
    max_date_use = min(max_parts, default=None)
    if (
        (min_date_use is not None)
        and (max_date_use is not None)
        and (min_date_use > max_date_use)
    ):
        raise _MinMaxDatePeriodError(min_date=min_date_use, max_date=max_date_use)
    return min_date_use, max_date_use


@dataclass(kw_only=True, slots=True)
class MinMaxDateError(Exception):
    min_date: Date
    max_date: Date


@dataclass(kw_only=True, slots=True)
class _MinMaxDatePeriodError(MinMaxDateError):
    @override
    def __str__(self) -> str:
        return (
            f"Min date must be at most max date; got {self.min_date} > {self.max_date}"
        )


##


class PeriodDict[T: Date | Time | ZonedDateTime | dt.date | dt.time | dt.datetime](
    TypedDict
):
    """A period as a dictionary."""

    start: T
    end: T


##


type _RoundDateDailyUnit = Literal["W", "D"]
type _RoundDateTimeUnit = Literal["H", "M", "S", "ms", "us", "ns"]
type _RoundDateOrDateTimeUnit = _RoundDateDailyUnit | _RoundDateTimeUnit


def round_date_or_date_time[T: Date | PlainDateTime | ZonedDateTime](
    date_or_date_time: T,
    delta: Delta,
    /,
    *,
    mode: DateTimeRoundMode = "half_even",
    weekday: Weekday | None = None,
) -> T:
    """Round a datetime."""
    increment, unit = _round_datetime_decompose(delta)
    match date_or_date_time, unit, weekday:
        case Date() as date, "W" | "D", _:
            return _round_date_weekly_or_daily(
                date, increment, unit, mode=mode, weekday=weekday
            )
        case Date() as date, "H" | "M" | "S" | "ms" | "us" | "ns", _:
            raise _RoundDateOrDateTimeDateWithIntradayDeltaError(date=date, delta=delta)
        case (PlainDateTime() | ZonedDateTime() as date_time, "W" | "D", _):
            return _round_date_time_weekly_or_daily(
                date_time, increment, unit, mode=mode, weekday=weekday
            )
        case (
            PlainDateTime() | ZonedDateTime() as date_time,
            "H" | "M" | "S" | "ms" | "us" | "ns",
            None,
        ):
            return _round_date_time_intraday(date_time, increment, unit, mode=mode)
        case (
            PlainDateTime() | ZonedDateTime() as date_time,
            "H" | "M" | "S" | "ms" | "us" | "ns",
            Weekday(),
        ):
            raise _RoundDateOrDateTimeDateTimeIntraDayWithWeekdayError(
                date_time=date_time, delta=delta, weekday=weekday
            )
        case never:
            assert_never(never)


def _round_datetime_decompose(delta: Delta, /) -> tuple[int, _RoundDateOrDateTimeUnit]:
    try:
        weeks = to_weeks(delta)
    except ToWeeksError:
        pass
    else:
        return weeks, "W"
    try:
        days = to_days(delta)
    except ToDaysError:
        pass
    else:
        return days, "D"
    try:
        hours = to_hours(delta)
    except ToHoursError:
        pass
    else:
        if (0 < hours < 24) and (24 % hours == 0):
            return hours, "H"
        raise _RoundDateOrDateTimeIncrementError(
            duration=delta, increment=hours, divisor=24
        )
    try:
        minutes = to_minutes(delta)
    except ToMinutesError:
        pass
    else:
        if (0 < minutes < 60) and (60 % minutes == 0):
            return minutes, "M"
        raise _RoundDateOrDateTimeIncrementError(
            duration=delta, increment=minutes, divisor=60
        )
    try:
        seconds = to_seconds(delta)
    except ToSecondsError:
        pass
    else:
        if (0 < seconds < 60) and (60 % seconds == 0):
            return seconds, "S"
        raise _RoundDateOrDateTimeIncrementError(
            duration=delta, increment=seconds, divisor=60
        )
    try:
        milliseconds = to_milliseconds(delta)
    except ToMillisecondsError:
        pass
    else:
        if (0 < milliseconds < 1000) and (1000 % milliseconds == 0):
            return milliseconds, "ms"
        raise _RoundDateOrDateTimeIncrementError(
            duration=delta, increment=milliseconds, divisor=1000
        )
    try:
        microseconds = to_microseconds(delta)
    except ToMicrosecondsError:
        pass
    else:
        if (0 < microseconds < 1000) and (1000 % microseconds == 0):
            return microseconds, "us"
        raise _RoundDateOrDateTimeIncrementError(
            duration=delta, increment=microseconds, divisor=1000
        )
    try:
        nanoseconds = to_nanoseconds(delta)
    except ToNanosecondsError:
        raise _RoundDateOrDateTimeInvalidDurationError(duration=delta) from None
    if (0 < nanoseconds < 1000) and (1000 % nanoseconds == 0):
        return nanoseconds, "ns"
    raise _RoundDateOrDateTimeIncrementError(
        duration=delta, increment=nanoseconds, divisor=1000
    )


def _round_date_weekly_or_daily(
    date: Date,
    increment: int,
    unit: _RoundDateDailyUnit,
    /,
    *,
    mode: DateTimeRoundMode = "half_even",
    weekday: Weekday | None = None,
) -> Date:
    match unit, weekday:
        case "W", _:
            return _round_date_weekly(date, increment, mode=mode, weekday=weekday)
        case "D", None:
            return _round_date_daily(date, increment, mode=mode)
        case "D", Weekday():
            raise _RoundDateOrDateTimeDateWithWeekdayError(weekday=weekday)
        case never:
            assert_never(never)


def _round_date_weekly(
    date: Date,
    increment: int,
    /,
    *,
    mode: DateTimeRoundMode = "half_even",
    weekday: Weekday | None = None,
) -> Date:
    mapping = {
        None: 0,
        Weekday.MONDAY: 0,
        Weekday.TUESDAY: 1,
        Weekday.WEDNESDAY: 2,
        Weekday.THURSDAY: 3,
        Weekday.FRIDAY: 4,
        Weekday.SATURDAY: 5,
        Weekday.SUNDAY: 6,
    }
    base = Date.MIN.add(days=mapping[weekday])
    return _round_date_daily(date, 7 * increment, mode=mode, base=base)


def _round_date_daily(
    date: Date,
    increment: int,
    /,
    *,
    mode: DateTimeRoundMode = "half_even",
    base: Date = Date.MIN,
) -> Date:
    quotient, remainder = divmod(date.days_since(base), increment)
    match mode:
        case "half_even":
            threshold = increment // 2 + (quotient % 2 == 0) or 1
        case "ceil":
            threshold = 1
        case "floor":
            threshold = increment + 1
        case "half_floor":
            threshold = increment // 2 + 1
        case "half_ceil":
            threshold = increment // 2 or 1
        case never:
            assert_never(never)
    round_up = remainder >= threshold
    return base.add(days=(quotient + round_up) * increment)


def _round_date_time_intraday[T: PlainDateTime | ZonedDateTime](
    date_time: T,
    increment: int,
    unit: _RoundDateTimeUnit,
    /,
    *,
    mode: DateTimeRoundMode = "half_even",
) -> T:
    match unit:
        case "H":
            unit_use = "hour"
        case "M":
            unit_use = "minute"
        case "S":
            unit_use = "second"
        case "ms":
            unit_use = "millisecond"
        case "us":
            unit_use = "microsecond"
        case "ns":
            unit_use = "nanosecond"
        case never:
            assert_never(never)
    return date_time.round(unit_use, increment=increment, mode=mode)


def _round_date_time_weekly_or_daily[T: PlainDateTime | ZonedDateTime](
    date_time: T,
    increment: int,
    unit: _RoundDateDailyUnit,
    /,
    *,
    mode: DateTimeRoundMode = "half_even",
    weekday: Weekday | None = None,
) -> T:
    rounded = cast("T", date_time.round("day", mode=mode))
    new_date = _round_date_weekly_or_daily(
        rounded.date(), increment, unit, mode=mode, weekday=weekday
    )
    return date_time.replace_date(new_date).replace_time(Time())


@dataclass(kw_only=True, slots=True)
class RoundDateOrDateTimeError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _RoundDateOrDateTimeIncrementError(RoundDateOrDateTimeError):
    duration: Delta
    increment: int
    divisor: int

    @override
    def __str__(self) -> str:
        return f"Duration {self.duration} increment must be a proper divisor of {self.divisor}; got {self.increment}"


@dataclass(kw_only=True, slots=True)
class _RoundDateOrDateTimeInvalidDurationError(RoundDateOrDateTimeError):
    duration: Delta

    @override
    def __str__(self) -> str:
        return f"Duration must be valid; got {self.duration}"


@dataclass(kw_only=True, slots=True)
class _RoundDateOrDateTimeDateWithIntradayDeltaError(RoundDateOrDateTimeError):
    date: Date
    delta: Delta

    @override
    def __str__(self) -> str:
        return f"Dates must not be given intraday durations; got {self.date} and {self.delta}"


@dataclass(kw_only=True, slots=True)
class _RoundDateOrDateTimeDateWithWeekdayError(RoundDateOrDateTimeError):
    weekday: Weekday

    @override
    def __str__(self) -> str:
        return f"Daily rounding must not be given a weekday; got {self.weekday}"


@dataclass(kw_only=True, slots=True)
class _RoundDateOrDateTimeDateTimeIntraDayWithWeekdayError(RoundDateOrDateTimeError):
    date_time: PlainDateTime | ZonedDateTime
    delta: Delta
    weekday: Weekday

    @override
    def __str__(self) -> str:
        return f"Date-times and intraday rounding must not be given a weekday; got {self.date_time}, {self.delta} and {self.weekday}"


##


@dataclass(repr=False, order=True, unsafe_hash=True, kw_only=False)
class TimePeriod:
    """A period of times."""

    start: Time
    end: Time

    @override
    def __repr__(self) -> str:
        cls = get_class_name(self)
        return f"{cls}({self.start}, {self.end})"

    def at(
        self, obj: Date | tuple[Date, Date], /, *, time_zone: TimeZoneLike = UTC
    ) -> ZonedDateTimePeriod:
        """Combine a date with a time to create a datetime."""
        match obj:
            case Date() as date:
                start = end = date
            case Date() as start, Date() as end:
                ...
            case never:
                assert_never(never)
        return DatePeriod(start, end).at((self.start, self.end), time_zone=time_zone)

    @classmethod
    def from_dict(cls, mapping: PeriodDict[Time] | PeriodDict[dt.time], /) -> Self:
        """Convert the dictionary to a period."""
        match mapping["start"]:
            case Time() as start:
                ...
            case dt.time() as py_time:
                start = Time.from_py_time(py_time)
            case never:
                assert_never(never)
        match mapping["end"]:
            case Time() as end:
                ...
            case dt.time() as py_time:
                end = Time.from_py_time(py_time)
            case never:
                assert_never(never)
        return cls(start=start, end=end)

    def replace(
        self, *, start: Time | Sentinel = sentinel, end: Time | Sentinel = sentinel
    ) -> Self:
        """Replace elements of the period."""
        return replace_non_sentinel(self, start=start, end=end)

    def to_dict(self) -> PeriodDict[Time]:
        """Convert the period to a dictionary."""
        return PeriodDict(start=self.start, end=self.end)

    def to_py_dict(self) -> PeriodDict[dt.time]:
        """Convert the period to a dictionary."""
        return PeriodDict(start=self.start.py_time(), end=self.end.py_time())


##


@overload
def to_date(date: Sentinel, /, *, time_zone: TimeZoneLike = UTC) -> Sentinel: ...
@overload
def to_date(
    date: MaybeCallableDateLike | None | dt.date = get_today,
    /,
    *,
    time_zone: TimeZoneLike = UTC,
) -> Date: ...
def to_date(
    date: MaybeCallableDateLike | dt.date | None | Sentinel = get_today,
    /,
    *,
    time_zone: TimeZoneLike = UTC,
) -> Date | Sentinel:
    """Convert to a date."""
    match date:
        case Date() | Sentinel():
            return date
        case None:
            return get_today(time_zone)
        case str():
            return Date.parse_iso(date)
        case dt.date():
            return Date.from_py_date(date)
        case Callable() as func:
            return to_date(func(), time_zone=time_zone)
        case never:
            assert_never(never)


##


def to_date_time_delta(nanos: int, /) -> DateTimeDelta:
    """Construct a date-time delta."""
    components = _to_time_delta_components(nanos)
    days, hours = divmod(components.hours, 24)
    weeks, days = divmod(days, 7)
    match sign(nanos):  # pragma: no cover
        case 1:
            if hours < 0:
                hours += 24
                days -= 1
            if days < 0:
                days += 7
                weeks -= 1
        case -1:
            if hours > 0:
                hours -= 24
                days += 1
            if days > 0:
                days -= 7
                weeks += 1
        case 0:
            ...
    return DateTimeDelta(
        weeks=weeks,
        days=days,
        hours=hours,
        minutes=components.minutes,
        seconds=components.seconds,
        microseconds=components.microseconds,
        milliseconds=components.milliseconds,
        nanoseconds=components.nanoseconds,
    )


##


def to_days(delta: Delta, /) -> int:
    """Compute the number of days in a delta."""
    match delta:
        case DateDelta():
            months, days = delta.in_months_days()
            if months != 0:
                raise _ToDaysMonthsError(delta=delta, months=months)
            return days
        case TimeDelta():
            nanos = to_nanoseconds(delta)
            days, remainder = divmod(nanos, 24 * 60 * 60 * int(1e9))
            if remainder != 0:
                raise _ToDaysNanosecondsError(delta=delta, nanoseconds=remainder)
            return days
        case DateTimeDelta():
            try:
                return to_days(delta.date_part()) + to_days(delta.time_part())
            except _ToDaysMonthsError as error:
                raise _ToDaysMonthsError(delta=delta, months=error.months) from None
            except _ToDaysNanosecondsError as error:
                raise _ToDaysNanosecondsError(
                    delta=delta, nanoseconds=error.nanoseconds
                ) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToDaysError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToDaysMonthsError(ToDaysError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToDaysNanosecondsError(ToDaysError):
    delta: TimeOrDateTimeDelta
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra nanoseconds; got {self.nanoseconds}"


##


def to_hours(delta: Delta, /) -> int:
    """Compute the number of hours in a delta."""
    match delta:
        case DateDelta():
            try:
                days = to_days(delta)
            except _ToDaysMonthsError as error:
                raise _ToHoursMonthsError(delta=delta, months=error.months) from None
            return 24 * days
        case TimeDelta():
            nanos = to_nanoseconds(delta)
            divisor = 60 * 60 * int(1e9)
            hours, remainder = divmod(nanos, divisor)
            if remainder != 0:
                raise _ToHoursNanosecondsError(delta=delta, nanoseconds=remainder)
            return hours
        case DateTimeDelta():
            try:
                return to_hours(delta.date_part()) + to_hours(delta.time_part())
            except _ToHoursMonthsError as error:
                raise _ToHoursMonthsError(delta=delta, months=error.months) from None
            except _ToHoursNanosecondsError as error:
                raise _ToHoursNanosecondsError(
                    delta=delta, nanoseconds=error.nanoseconds
                ) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToHoursError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToHoursMonthsError(ToHoursError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToHoursNanosecondsError(ToHoursError):
    delta: TimeOrDateTimeDelta
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra nanoseconds; got {self.nanoseconds}"


##


def to_microseconds(delta: Delta, /) -> int:
    """Compute the number of microseconds in a delta."""
    match delta:
        case DateDelta():
            try:
                days = to_days(delta)
            except _ToDaysMonthsError as error:
                raise _ToMicrosecondsMonthsError(
                    delta=delta, months=error.months
                ) from None
            return 24 * 60 * 60 * int(1e6) * days
        case TimeDelta():
            nanos = to_nanoseconds(delta)
            microseconds, remainder = divmod(nanos, int(1e3))
            if remainder != 0:
                raise _ToMicrosecondsNanosecondsError(
                    delta=delta, nanoseconds=remainder
                )
            return microseconds
        case DateTimeDelta():
            try:
                return to_microseconds(delta.date_part()) + to_microseconds(
                    delta.time_part()
                )
            except _ToMicrosecondsMonthsError as error:
                raise _ToMicrosecondsMonthsError(
                    delta=delta, months=error.months
                ) from None
            except _ToMicrosecondsNanosecondsError as error:
                raise _ToMicrosecondsNanosecondsError(
                    delta=delta, nanoseconds=error.nanoseconds
                ) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToMicrosecondsError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToMicrosecondsMonthsError(ToMicrosecondsError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToMicrosecondsNanosecondsError(ToMicrosecondsError):
    delta: TimeOrDateTimeDelta
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra nanoseconds; got {self.nanoseconds}"


##


def to_milliseconds(delta: Delta, /) -> int:
    """Compute the number of milliseconds in a delta."""
    match delta:
        case DateDelta():
            try:
                days = to_days(delta)
            except _ToDaysMonthsError as error:
                raise _ToMillisecondsMonthsError(
                    delta=delta, months=error.months
                ) from None
            return 24 * 60 * 60 * int(1e3) * days
        case TimeDelta():
            nanos = to_nanoseconds(delta)
            milliseconds, remainder = divmod(nanos, int(1e6))
            if remainder != 0:
                raise _ToMillisecondsNanosecondsError(
                    delta=delta, nanoseconds=remainder
                )
            return milliseconds
        case DateTimeDelta():
            try:
                return to_milliseconds(delta.date_part()) + to_milliseconds(
                    delta.time_part()
                )
            except _ToMillisecondsMonthsError as error:
                raise _ToMillisecondsMonthsError(
                    delta=delta, months=error.months
                ) from None
            except _ToMillisecondsNanosecondsError as error:
                raise _ToMillisecondsNanosecondsError(
                    delta=delta, nanoseconds=error.nanoseconds
                ) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToMillisecondsError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToMillisecondsMonthsError(ToMillisecondsError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToMillisecondsNanosecondsError(ToMillisecondsError):
    delta: TimeOrDateTimeDelta
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra nanoseconds; got {self.nanoseconds}"


##


def to_minutes(delta: Delta, /) -> int:
    """Compute the number of minutes in a delta."""
    match delta:
        case DateDelta():
            try:
                days = to_days(delta)
            except _ToDaysMonthsError as error:
                raise _ToMinutesMonthsError(delta=delta, months=error.months) from None
            return 24 * 60 * days
        case TimeDelta():
            nanos = to_nanoseconds(delta)
            minutes, remainder = divmod(nanos, 60 * int(1e9))
            if remainder != 0:
                raise _ToMinutesNanosecondsError(delta=delta, nanoseconds=remainder)
            return minutes
        case DateTimeDelta():
            try:
                return to_minutes(delta.date_part()) + to_minutes(delta.time_part())
            except _ToMinutesMonthsError as error:
                raise _ToMinutesMonthsError(delta=delta, months=error.months) from None
            except _ToMinutesNanosecondsError as error:
                raise _ToMinutesNanosecondsError(
                    delta=delta, nanoseconds=error.nanoseconds
                ) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToMinutesError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToMinutesMonthsError(ToMinutesError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToMinutesNanosecondsError(ToMinutesError):
    delta: TimeOrDateTimeDelta
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra nanoseconds; got {self.nanoseconds}"


##


def to_months(delta: DateOrDateTimeDelta, /) -> int:
    """Compute the number of months in a delta."""
    match delta:
        case DateDelta():
            months, days = delta.in_months_days()
            if days != 0:
                raise _ToMonthsDaysError(delta=delta, days=days)
            return months
        case DateTimeDelta():
            if delta.time_part() != TimeDelta():
                raise _ToMonthsTimeError(delta=delta)
            try:
                return to_months(delta.date_part())
            except _ToMonthsDaysError as error:
                raise _ToMonthsDaysError(delta=delta, days=error.days) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToMonthsError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToMonthsDaysError(ToMonthsError):
    delta: DateOrDateTimeDelta
    days: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain days; got {self.days}"


@dataclass(kw_only=True, slots=True)
class _ToMonthsTimeError(ToMonthsError):
    delta: DateTimeDelta

    @override
    def __str__(self) -> str:
        return f"Delta must not contain a time part; got {self.delta.time_part()}"


##


def to_months_and_days(delta: DateOrDateTimeDelta, /) -> tuple[int, int]:
    """Compute the number of months & days in a delta."""
    match delta:
        case DateDelta():
            return delta.in_months_days()
        case DateTimeDelta():
            if delta.time_part() != TimeDelta():
                raise ToMonthsAndDaysError(delta=delta)
            return to_months_and_days(delta.date_part())
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToMonthsAndDaysError(Exception):
    delta: DateTimeDelta

    @override
    def __str__(self) -> str:
        return f"Delta must not contain a time part; got {self.delta.time_part()}"


##


def to_nanoseconds(delta: Delta, /) -> int:
    """Compute the number of nanoseconds in a date-time delta."""
    match delta:
        case DateDelta():
            try:
                days = to_days(delta)
            except _ToDaysMonthsError as error:
                raise ToNanosecondsError(delta=delta, months=error.months) from None
            return 24 * 60 * 60 * int(1e9) * days
        case TimeDelta():
            return delta.in_nanoseconds()
        case DateTimeDelta():
            try:
                return to_nanoseconds(delta.date_part()) + to_nanoseconds(
                    delta.time_part()
                )
            except ToNanosecondsError as error:
                raise ToNanosecondsError(delta=delta, months=error.months) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToNanosecondsError(Exception):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


##


@overload
def to_py_date_or_date_time(date_or_date_time: Date, /) -> dt.date: ...
@overload
def to_py_date_or_date_time(date_or_date_time: ZonedDateTime, /) -> dt.datetime: ...
@overload
def to_py_date_or_date_time(date_or_date_time: None, /) -> None: ...
def to_py_date_or_date_time(
    date_or_date_time: Date | ZonedDateTime | None, /
) -> dt.date | None:
    """Convert a Date or ZonedDateTime into a standard library equivalent."""
    match date_or_date_time:
        case Date() as date:
            return date.py_date()
        case ZonedDateTime() as date_time:
            return date_time.py_datetime()
        case None:
            return None
        case never:
            assert_never(never)


##


@overload
def to_py_time_delta(delta: Delta, /) -> dt.timedelta: ...
@overload
def to_py_time_delta(delta: None, /) -> None: ...
def to_py_time_delta(delta: Delta | None, /) -> dt.timedelta | None:
    """Try convert a DateDelta to a standard library timedelta."""
    match delta:
        case DateDelta():
            return dt.timedelta(days=to_days(delta))
        case TimeDelta():
            nanos = delta.in_nanoseconds()
            micros, remainder = divmod(nanos, 1000)
            if remainder != 0:
                raise ToPyTimeDeltaError(nanoseconds=remainder)
            return dt.timedelta(microseconds=micros)
        case DateTimeDelta():
            return to_py_time_delta(delta.date_part()) + to_py_time_delta(
                delta.time_part()
            )
        case None:
            return None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToPyTimeDeltaError(Exception):
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Time delta must not contain nanoseconds; got {self.nanoseconds}"


##


def to_seconds(delta: Delta, /) -> int:
    """Compute the number of seconds in a delta."""
    match delta:
        case DateDelta():
            try:
                days = to_days(delta)
            except _ToDaysMonthsError as error:
                raise _ToSecondsMonthsError(delta=delta, months=error.months) from None
            return 24 * 60 * 60 * days
        case TimeDelta():
            nanos = to_nanoseconds(delta)
            seconds, remainder = divmod(nanos, int(1e9))
            if remainder != 0:
                raise _ToSecondsNanosecondsError(delta=delta, nanoseconds=remainder)
            return seconds
        case DateTimeDelta():
            try:
                return to_seconds(delta.date_part()) + to_seconds(delta.time_part())
            except _ToSecondsMonthsError as error:
                raise _ToSecondsMonthsError(delta=delta, months=error.months) from None
            except _ToSecondsNanosecondsError as error:
                raise _ToSecondsNanosecondsError(
                    delta=delta, nanoseconds=error.nanoseconds
                ) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToSecondsError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToSecondsMonthsError(ToSecondsError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToSecondsNanosecondsError(ToSecondsError):
    delta: TimeOrDateTimeDelta
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra nanoseconds; got {self.nanoseconds}"


##


@overload
def to_time(time: Sentinel, /, *, time_zone: TimeZoneLike = UTC) -> Sentinel: ...
@overload
def to_time(
    time: MaybeCallableTimeLike | None | dt.time = get_time,
    /,
    *,
    time_zone: TimeZoneLike = UTC,
) -> Time: ...
def to_time(
    time: MaybeCallableTimeLike | dt.time | None | Sentinel = get_time,
    /,
    *,
    time_zone: TimeZoneLike = UTC,
) -> Time | Sentinel:
    """Convert to a time."""
    match time:
        case Time() | Sentinel():
            return time
        case None:
            return get_time(time_zone)
        case str():
            return Time.parse_iso(time)
        case dt.time():
            return Time.from_py_time(time)
        case Callable() as func:
            return to_time(func(), time_zone=time_zone)
        case never:
            assert_never(never)


##


def to_time_delta(nanos: int, /) -> TimeDelta:
    """Construct a time delta."""
    components = _to_time_delta_components(nanos)
    return TimeDelta(
        hours=components.hours,
        minutes=components.minutes,
        seconds=components.seconds,
        microseconds=components.microseconds,
        milliseconds=components.milliseconds,
        nanoseconds=components.nanoseconds,
    )


@dataclass(kw_only=True, slots=True)
class _TimeDeltaComponents:
    hours: int
    minutes: int
    seconds: int
    microseconds: int
    milliseconds: int
    nanoseconds: int


def _to_time_delta_components(nanos: int, /) -> _TimeDeltaComponents:
    sign_use = sign(nanos)
    micros, nanos = divmod(nanos, int(1e3))
    millis, micros = divmod(micros, int(1e3))
    secs, millis = divmod(millis, int(1e3))
    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    match sign_use:  # pragma: no cover
        case 1:
            if nanos < 0:
                nanos += int(1e3)
                micros -= 1
            if micros < 0:
                micros += int(1e3)
                millis -= 1
            if millis < 0:
                millis += int(1e3)
                secs -= 1
            if secs < 0:
                secs += 60
                mins -= 1
            if mins < 0:
                mins += 60
                hours -= 1
        case -1:
            if nanos > 0:
                nanos -= int(1e3)
                micros += 1
            if micros > 0:
                micros -= int(1e3)
                millis += 1
            if millis > 0:
                millis -= int(1e3)
                secs += 1
            if secs > 0:
                secs -= 60
                mins += 1
            if mins > 0:
                mins -= 60
                hours += 1
        case 0:
            ...
    return _TimeDeltaComponents(
        hours=hours,
        minutes=mins,
        seconds=secs,
        microseconds=micros,
        milliseconds=millis,
        nanoseconds=nanos,
    )


##


def to_weeks(delta: Delta, /) -> int:
    """Compute the number of weeks in a delta."""
    try:
        days = to_days(delta)
    except _ToDaysMonthsError as error:
        raise _ToWeeksMonthsError(delta=error.delta, months=error.months) from None
    except _ToDaysNanosecondsError as error:
        raise _ToWeeksNanosecondsError(
            delta=error.delta, nanoseconds=error.nanoseconds
        ) from None
    weeks, remainder = divmod(days, 7)
    if remainder != 0:
        raise _ToWeeksDaysError(delta=delta, days=remainder) from None
    return weeks


@dataclass(kw_only=True, slots=True)
class ToWeeksError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToWeeksMonthsError(ToWeeksError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToWeeksNanosecondsError(ToWeeksError):
    delta: TimeOrDateTimeDelta
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra nanoseconds; got {self.nanoseconds}"


@dataclass(kw_only=True, slots=True)
class _ToWeeksDaysError(ToWeeksError):
    delta: Delta
    days: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain extra days; got {self.days}"


##


def to_years(delta: DateOrDateTimeDelta, /) -> int:
    """Compute the number of years in a delta."""
    match delta:
        case DateDelta():
            years, months, days = delta.in_years_months_days()
            if months != 0:
                raise _ToYearsMonthsError(delta=delta, months=months)
            if days != 0:
                raise _ToYearsDaysError(delta=delta, days=days)
            return years
        case DateTimeDelta():
            if delta.time_part() != TimeDelta():
                raise _ToYearsTimeError(delta=delta)
            try:
                return to_years(delta.date_part())
            except _ToYearsMonthsError as error:
                raise _ToYearsMonthsError(delta=delta, months=error.months) from None
            except _ToYearsDaysError as error:
                raise _ToYearsDaysError(delta=delta, days=error.days) from None
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToYearsError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToYearsMonthsError(ToYearsError):
    delta: DateOrDateTimeDelta
    months: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain months; got {self.months}"


@dataclass(kw_only=True, slots=True)
class _ToYearsDaysError(ToYearsError):
    delta: DateOrDateTimeDelta
    days: int

    @override
    def __str__(self) -> str:
        return f"Delta must not contain days; got {self.days}"


@dataclass(kw_only=True, slots=True)
class _ToYearsTimeError(ToYearsError):
    delta: DateTimeDelta

    @override
    def __str__(self) -> str:
        return f"Delta must not contain a time part; got {self.delta.time_part()}"


##


@overload
def to_zoned_date_time(
    date_time: Sentinel, /, *, time_zone: TimeZoneLike | None = None
) -> Sentinel: ...
@overload
def to_zoned_date_time(
    date_time: MaybeCallableZonedDateTimeLike | dt.datetime | None = get_now,
    /,
    *,
    time_zone: TimeZoneLike | None = None,
) -> ZonedDateTime: ...
def to_zoned_date_time(
    date_time: MaybeCallableZonedDateTimeLike | dt.datetime | None | Sentinel = get_now,
    /,
    *,
    time_zone: TimeZoneLike | None = None,
) -> ZonedDateTime | Sentinel:
    """Convert to a zoned date-time."""
    match date_time:
        case ZonedDateTime() as date_time_use:
            ...
        case Sentinel():
            return sentinel
        case None:
            return get_now(UTC if time_zone is None else time_zone)
        case str() as text:
            date_time_use = ZonedDateTime.parse_iso(text.replace("~", "/"))
        case dt.datetime() as py_date_time:
            if isinstance(date_time.tzinfo, ZoneInfo):
                py_date_time_use = py_date_time
            elif date_time.tzinfo is dt.UTC:
                py_date_time_use = py_date_time.astimezone(UTC)
            else:
                raise ToZonedDateTimeError(date_time=date_time)
            date_time_use = ZonedDateTime.from_py_datetime(py_date_time_use)
        case Callable() as func:
            return to_zoned_date_time(func(), time_zone=time_zone)
        case never:
            assert_never(never)
    if time_zone is None:
        return date_time_use
    return date_time_use.to_tz(to_time_zone_name(time_zone))


@dataclass(kw_only=True, slots=True)
class ToZonedDateTimeError(Exception):
    date_time: dt.datetime

    @override
    def __str__(self) -> str:
        return f"Expected date-time to have a `ZoneInfo` or `dt.UTC` as its timezone; got {self.date_time.tzinfo}"


##


def two_digit_year_month(year: int, month: int, /) -> YearMonth:
    """Construct a year-month from a 2-digit year."""
    min_year = DATE_TWO_DIGIT_YEAR_MIN.year
    max_year = DATE_TWO_DIGIT_YEAR_MAX.year
    years = range(min_year, max_year + 1)
    (year_use,) = (y for y in years if y % 100 == year)
    return YearMonth(year_use, month)


##


class WheneverLogRecord(LogRecord):
    """Log record powered by `whenever`."""

    zoned_datetime: str

    @override
    def __init__(
        self,
        name: str,
        level: int,
        pathname: str,
        lineno: int,
        msg: object,
        args: Any,
        exc_info: Any,
        func: str | None = None,
        sinfo: str | None = None,
    ) -> None:
        super().__init__(
            name, level, pathname, lineno, msg, args, exc_info, func, sinfo
        )
        length = self._get_length()
        plain = format(get_now_local().to_plain().format_iso(), f"{length}s")
        self.zoned_datetime = f"{plain}[{LOCAL_TIME_ZONE_NAME}]"

    @classmethod
    @cache
    def _get_length(cls) -> int:
        """Get maximum length of a formatted string."""
        now = get_now_local().replace(nanosecond=1000).to_plain()
        return len(now.format_iso())


##


@dataclass(repr=False, order=True, unsafe_hash=True, kw_only=False)
class ZonedDateTimePeriod:
    """A period of time."""

    start: ZonedDateTime
    end: ZonedDateTime

    def __post_init__(self) -> None:
        if self.start > self.end:
            raise _ZonedDateTimePeriodInvalidError(start=self.start, end=self.end)
        if self.start.tz != self.end.tz:
            raise _ZonedDateTimePeriodTimeZoneError(
                start=ZoneInfo(self.start.tz), end=ZoneInfo(self.end.tz)
            )

    def __add__(self, other: TimeDelta, /) -> Self:
        """Offset the period."""
        return self.replace(start=self.start + other, end=self.end + other)

    def __contains__(self, other: ZonedDateTime, /) -> bool:
        """Check if a date/datetime lies in the period."""
        return self.start <= other <= self.end

    @override
    def __repr__(self) -> str:
        cls = get_class_name(self)
        return f"{cls}({self.start.to_plain()}, {self.end.to_plain()}[{self.time_zone.key}])"

    def __sub__(self, other: TimeDelta, /) -> Self:
        """Offset the period."""
        return self.replace(start=self.start - other, end=self.end - other)

    @property
    def delta(self) -> TimeDelta:
        """The duration of the period."""
        return self.end - self.start

    @overload
    def exact_eq(self, period: ZonedDateTimePeriod, /) -> bool: ...
    @overload
    def exact_eq(self, start: ZonedDateTime, end: ZonedDateTime, /) -> bool: ...
    @overload
    def exact_eq(
        self, start: PlainDateTime, end: PlainDateTime, time_zone: ZoneInfo, /
    ) -> bool: ...
    def exact_eq(self, *args: Any) -> bool:
        """Check if a period is exactly equal to another."""
        if (len(args) == 1) and isinstance(args[0], ZonedDateTimePeriod):
            return self.start.exact_eq(args[0].start) and self.end.exact_eq(args[0].end)
        if (
            (len(args) == 2)
            and isinstance(args[0], ZonedDateTime)
            and isinstance(args[1], ZonedDateTime)
        ):
            return self.exact_eq(ZonedDateTimePeriod(args[0], args[1]))
        if (
            (len(args) == 3)
            and isinstance(args[0], PlainDateTime)
            and isinstance(args[1], PlainDateTime)
            and isinstance(args[2], ZoneInfo)
        ):
            return self.exact_eq(
                ZonedDateTimePeriod(
                    args[0].assume_tz(args[2].key), args[1].assume_tz(args[2].key)
                )
            )
        raise _ZonedDateTimePeriodExactEqError(args=args)

    def format_compact(self) -> str:
        """Format the period in a compact fashion."""
        fc, start, end = format_compact, self.start, self.end
        if start == end:
            if end.second != 0:
                return f"{fc(start)}="
            if end.minute != 0:
                return f"{fc(start, fmt='%Y%m%dT%H%M')}="
            return f"{fc(start, fmt='%Y%m%dT%H')}="
        if start.date() == end.date():
            if end.second != 0:
                return f"{fc(start.to_plain())}-{fc(end, fmt='%H%M%S')}"
            if end.minute != 0:
                return f"{fc(start.to_plain())}-{fc(end, fmt='%H%M')}"
            return f"{fc(start.to_plain())}-{fc(end, fmt='%H')}"
        if start.date().year_month() == end.date().year_month():
            if end.second != 0:
                return f"{fc(start.to_plain())}-{fc(end, fmt='%dT%H%M%S')}"
            if end.minute != 0:
                return f"{fc(start.to_plain())}-{fc(end, fmt='%dT%H%M')}"
            return f"{fc(start.to_plain())}-{fc(end, fmt='%dT%H')}"
        if start.year == end.year:
            if end.second != 0:
                return f"{fc(start.to_plain())}-{fc(end, fmt='%m%dT%H%M%S')}"
            if end.minute != 0:
                return f"{fc(start.to_plain())}-{fc(end, fmt='%m%dT%H%M')}"
            return f"{fc(start.to_plain())}-{fc(end, fmt='%m%dT%H')}"
        if end.second != 0:
            return f"{fc(start.to_plain())}-{fc(end)}"
        if end.minute != 0:
            return f"{fc(start.to_plain())}-{fc(end, fmt='%Y%m%dT%H%M')}"
        return f"{fc(start.to_plain())}-{fc(end, fmt='%Y%m%dT%H')}"

    @classmethod
    def from_dict(
        cls, mapping: PeriodDict[ZonedDateTime] | PeriodDict[dt.datetime], /
    ) -> Self:
        """Convert the dictionary to a period."""
        match mapping["start"]:
            case ZonedDateTime() as start:
                ...
            case dt.date() as py_datetime:
                start = ZonedDateTime.from_py_datetime(py_datetime)
            case never:
                assert_never(never)
        match mapping["end"]:
            case ZonedDateTime() as end:
                ...
            case dt.date() as py_datetime:
                end = ZonedDateTime.from_py_datetime(py_datetime)
            case never:
                assert_never(never)
        return cls(start=start, end=end)

    def replace(
        self,
        *,
        start: ZonedDateTime | Sentinel = sentinel,
        end: ZonedDateTime | Sentinel = sentinel,
    ) -> Self:
        """Replace elements of the period."""
        return replace_non_sentinel(self, start=start, end=end)

    @property
    def time_zone(self) -> ZoneInfo:
        """The time zone of the period."""
        return ZoneInfo(self.start.tz)

    def to_dict(self) -> PeriodDict[ZonedDateTime]:
        """Convert the period to a dictionary."""
        return PeriodDict(start=self.start, end=self.end)

    def to_py_dict(self) -> PeriodDict[dt.datetime]:
        """Convert the period to a dictionary."""
        return PeriodDict(start=self.start.py_datetime(), end=self.end.py_datetime())

    def to_tz(self, time_zone: TimeZoneLike, /) -> Self:
        """Convert the time zone."""
        tz = to_time_zone_name(time_zone)
        return self.replace(start=self.start.to_tz(tz), end=self.end.to_tz(tz))


@dataclass(kw_only=True, slots=True)
class ZonedDateTimePeriodError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ZonedDateTimePeriodInvalidError[T: Date | ZonedDateTime](
    ZonedDateTimePeriodError
):
    start: T
    end: T

    @override
    def __str__(self) -> str:
        return f"Invalid period; got {self.start} > {self.end}"


@dataclass(kw_only=True, slots=True)
class _ZonedDateTimePeriodTimeZoneError(ZonedDateTimePeriodError):
    start: ZoneInfo
    end: ZoneInfo

    @override
    def __str__(self) -> str:
        return f"Period must contain exactly one time zone; got {self.start} and {self.end}"


@dataclass(kw_only=True, slots=True)
class _ZonedDateTimePeriodExactEqError(ZonedDateTimePeriodError):
    args: tuple[Any, ...]

    @override
    def __str__(self) -> str:
        return f"Invalid arguments; got {self.args}"


__all__ = [
    "DATE_DELTA_MAX",
    "DATE_DELTA_MIN",
    "DATE_DELTA_PARSABLE_MAX",
    "DATE_DELTA_PARSABLE_MIN",
    "DATE_TIME_DELTA_MAX",
    "DATE_TIME_DELTA_MIN",
    "DATE_TIME_DELTA_PARSABLE_MAX",
    "DATE_TIME_DELTA_PARSABLE_MIN",
    "DATE_TWO_DIGIT_YEAR_MAX",
    "DATE_TWO_DIGIT_YEAR_MIN",
    "DAY",
    "HOUR",
    "MICROSECOND",
    "MILLISECOND",
    "MINUTE",
    "MONTH",
    "NOW_LOCAL",
    "NOW_LOCAL_PLAIN",
    "NOW_PLAIN",
    "SECOND",
    "TIME_DELTA_MAX",
    "TIME_DELTA_MIN",
    "TIME_LOCAL",
    "TIME_UTC",
    "TODAY_LOCAL",
    "TODAY_UTC",
    "WEEK",
    "YEAR",
    "ZERO_DAYS",
    "ZERO_TIME",
    "ZONED_DATE_TIME_MAX",
    "ZONED_DATE_TIME_MIN",
    "DatePeriod",
    "DatePeriodError",
    "MeanDateTimeError",
    "MinMaxDateError",
    "PeriodDict",
    "RoundDateOrDateTimeError",
    "TimePeriod",
    "ToDaysError",
    "ToMinutesError",
    "ToMonthsAndDaysError",
    "ToMonthsError",
    "ToNanosecondsError",
    "ToPyTimeDeltaError",
    "ToSecondsError",
    "ToWeeksError",
    "ToYearsError",
    "WheneverLogRecord",
    "ZonedDateTimePeriod",
    "ZonedDateTimePeriodError",
    "add_year_month",
    "datetime_utc",
    "diff_year_month",
    "format_compact",
    "from_timestamp",
    "from_timestamp_millis",
    "from_timestamp_nanos",
    "get_now",
    "get_now_local",
    "get_now_local_plain",
    "get_now_plain",
    "get_time",
    "get_time_local",
    "get_today",
    "get_today_local",
    "is_weekend",
    "mean_datetime",
    "min_max_date",
    "round_date_or_date_time",
    "sub_year_month",
    "to_date",
    "to_date_time_delta",
    "to_days",
    "to_microseconds",
    "to_milliseconds",
    "to_minutes",
    "to_months",
    "to_months_and_days",
    "to_nanoseconds",
    "to_py_date_or_date_time",
    "to_py_time_delta",
    "to_seconds",
    "to_time",
    "to_weeks",
    "to_years",
    "to_zoned_date_time",
    "two_digit_year_month",
]



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Literal, overload

from utilities.errors import ImpossibleCaseError
from utilities.subprocess import (
    MKTEMP_DIR_CMD,
    maybe_sudo_cmd,
    mkdir,
    mkdir_cmd,
    rm_cmd,
    run,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from utilities.types import PathLike, StrStrMapping


@overload
def docker_cp(
    src: tuple[str, PathLike], dest: PathLike, /, *, sudo: bool = False
) -> None: ...
@overload
def docker_cp(
    src: PathLike, dest: tuple[str, PathLike], /, *, sudo: bool = False
) -> None: ...
def docker_cp(
    src: PathLike | tuple[str, PathLike],
    dest: PathLike | tuple[str, PathLike],
    /,
    *,
    sudo: bool = False,
) -> None:
    match src, dest:
        case Path() | str(), (str() as cont, Path() | str() as dest_path):
            docker_exec(
                cont, *maybe_sudo_cmd(*mkdir_cmd(dest_path, parent=True), sudo=sudo)
            )
            run(*docker_cp_cmd(src, dest, sudo=sudo))
        case (str(), Path() | str()), Path() | str():
            mkdir(dest, parent=True, sudo=sudo)
            run(*docker_cp_cmd(src, dest, sudo=sudo))
        case _:  # pragma: no cover
            raise ImpossibleCaseError(case=[f"{src}", f"{dest=}"])


@overload
def docker_cp_cmd(
    src: tuple[str, PathLike], dest: PathLike, /, *, sudo: bool = False
) -> list[str]: ...
@overload
def docker_cp_cmd(
    src: PathLike, dest: tuple[str, PathLike], /, *, sudo: bool = False
) -> list[str]: ...
def docker_cp_cmd(
    src: PathLike | tuple[str, PathLike],
    dest: PathLike | tuple[str, PathLike],
    /,
    *,
    sudo: bool = False,
) -> list[str]:
    match src, dest:
        case (Path() | str()) as src_use, (
            str() as dest_cont,
            Path() | str() as dest_path,
        ):
            dest_use = f"{dest_cont}:{dest_path}"
        case (str() as src_cont, (Path() | str()) as src_path), (
            Path() | str() as dest_use
        ):
            src_use = f"{src_cont}:{src_path}"
        case _:  # pragma: no cover
            raise ImpossibleCaseError(case=[f"{src}", f"{dest=}"])
    parts: list[str] = ["docker", "cp", str(src_use), str(dest_use)]
    return maybe_sudo_cmd(*parts, sudo=sudo)


@overload
def docker_exec(
    container: str,
    cmd: str,
    /,
    *args: str,
    env: StrStrMapping | None = None,
    user: str | None = None,
    workdir: PathLike | None = None,
    shell: bool = False,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: Literal[True],
    return_stdout: bool = False,
    return_stderr: bool = False,
    **env_kwargs: str,
) -> str: ...
@overload
def docker_exec(
    container: str,
    cmd: str,
    /,
    *args: str,
    env: StrStrMapping | None = None,
    user: str | None = None,
    workdir: PathLike | None = None,
    shell: bool = False,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: Literal[True],
    return_stderr: bool = False,
    **env_kwargs: str,
) -> str: ...
@overload
def docker_exec(
    container: str,
    cmd: str,
    /,
    *args: str,
    env: StrStrMapping | None = None,
    user: str | None = None,
    workdir: PathLike | None = None,
    shell: bool = False,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: bool = False,
    return_stderr: Literal[True],
    **env_kwargs: str,
) -> str: ...
@overload
def docker_exec(
    container: str,
    cmd: str,
    /,
    *args: str,
    env: StrStrMapping | None = None,
    user: str | None = None,
    workdir: PathLike | None = None,
    shell: bool = False,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: Literal[False] = False,
    return_stdout: Literal[False] = False,
    return_stderr: Literal[False] = False,
    **env_kwargs: str,
) -> None: ...
@overload
def docker_exec(
    container: str,
    cmd: str,
    /,
    *args: str,
    env: StrStrMapping | None = None,
    user: str | None = None,
    workdir: PathLike | None = None,
    shell: bool = False,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: bool = False,
    return_stderr: bool = False,
    **env_kwargs: str,
) -> str | None: ...
def docker_exec(
    container: str,
    cmd: str,
    /,
    *args: str,
    env: StrStrMapping | None = None,
    user: str | None = None,
    workdir: PathLike | None = None,
    shell: bool = False,
    print: bool = False,  # noqa: A002
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: bool = False,
    return_stderr: bool = False,
    **env_kwargs: str,
) -> str | None:
    cmd_use = docker_exec_cmd(  # skipif-ci
        container, cmd, *args, env=env, user=user, workdir=workdir, **env_kwargs
    )
    if shell:  # skipif-ci
        return run(
            " ".join(cmd_use),
            shell=shell,
            print=print,
            print_stdout=print_stdout,
            print_stderr=print_stderr,
            return_=return_,
            return_stdout=return_stdout,
            return_stderr=return_stderr,
        )
    return run(  # skipif-ci
        *cmd_use,
        shell=shell,
        print=print,
        print_stdout=print_stdout,
        print_stderr=print_stderr,
        return_=return_,
        return_stdout=return_stdout,
        return_stderr=return_stderr,
    )


def docker_exec_cmd(
    container: str,
    cmd: str,
    /,
    *args: str,
    env: StrStrMapping | None = None,
    user: str | None = None,
    workdir: PathLike | None = None,
    **env_kwargs: str,
) -> list[str]:
    """Build a command for `docker exec`."""
    parts: list[str] = ["docker", "exec"]
    mapping: dict[str, str] = ({} if env is None else dict(env)) | env_kwargs
    for key, value in mapping.items():
        parts.extend(["--env", f"{key}={value}"])
    if user is not None:
        parts.extend(["--user", user])
    if workdir is not None:
        parts.extend(["--workdir", str(workdir)])
    return [*parts, container, cmd, *args]


@contextmanager
def yield_docker_temp_dir(
    container: str, /, *, user: str | None = None
) -> Iterator[Path]:
    path = Path(  # skipif-ci
        docker_exec(container, *MKTEMP_DIR_CMD, user=user, return_=True).rstrip("\n")
    )
    try:  # skipif-ci
        yield path
    finally:  # skipif-ci
        docker_exec(container, *rm_cmd(path), user=user)


__all__ = ["docker_cp_cmd", "docker_exec", "docker_exec_cmd", "yield_docker_temp_dir"]



from __future__ import annotations

import re
from collections import deque
from collections.abc import Callable
from dataclasses import dataclass
from itertools import chain
from os import getpid
from re import IGNORECASE, VERBOSE, escape, search
from textwrap import dedent
from threading import get_ident
from time import time_ns
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    assert_never,
    overload,
    override,
)
from uuid import uuid4

from utilities.iterables import CheckDuplicatesError, check_duplicates, transpose
from utilities.reprlib import get_repr
from utilities.sentinel import Sentinel

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence

    from utilities.types import MaybeCallableBoolLike, MaybeCallableStr, StrStrMapping


_DEFAULT_SEPARATOR = ","


##


def kebab_case(text: str, /) -> str:
    """Convert text into kebab case."""
    return _kebab_snake_case(text, "-")


##


def parse_bool(text: str, /) -> bool:
    """Parse text into a boolean value."""
    if search(r"^(0|False|N|No|Off)$", text, flags=IGNORECASE):
        return False
    if search(r"^(1|True|Y|Yes|On)$", text, flags=IGNORECASE):
        return True
    raise ParseBoolError(text=text)


@dataclass(kw_only=True, slots=True)
class ParseBoolError(Exception):
    text: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse boolean value; got {self.text!r}"


##


def parse_none(text: str, /) -> None:
    """Parse text into the None value."""
    if search(r"^(|None)$", text, flags=IGNORECASE):
        return
    raise ParseNoneError(text=text)


@dataclass(kw_only=True, slots=True)
class ParseNoneError(Exception):
    text: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse null value; got {self.text!r}"


##


def pascal_case(text: str, /) -> str:
    """Convert text to pascal case."""
    parts = _SPLIT_TEXT.findall(text)
    parts = [p for p in parts if len(p) >= 1]
    parts = list(map(_pascal_case_one, parts))
    return "".join(parts)


def _pascal_case_one(text: str, /) -> str:
    return text if text.isupper() else text.title()


##


def prompt_bool(prompt: object = "", /, *, confirm: bool = False) -> bool:
    """Prompt for a boolean."""
    return True if confirm else parse_bool(input(prompt))


##


def repr_encode(obj: Any, /) -> bytes:
    """Return the representation of the object encoded as bytes."""
    return repr(obj).encode()


##


def snake_case(text: str, /) -> str:
    """Convert text into snake case."""
    return _kebab_snake_case(text, "_")


##


LIST_SEPARATOR = _DEFAULT_SEPARATOR
PAIR_SEPARATOR = "="
BRACKETS = [("(", ")"), ("[", "]"), ("{", "}")]


@overload
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = _DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: Literal[True],
) -> StrStrMapping: ...
@overload
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = _DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: Literal[False] = False,
) -> Sequence[tuple[str, str]]: ...
@overload
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = _DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: bool = False,
) -> Sequence[tuple[str, str]] | StrStrMapping: ...
def split_key_value_pairs(
    text: str,
    /,
    *,
    list_separator: str = _DEFAULT_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    mapping: bool = False,
) -> Sequence[tuple[str, str]] | StrStrMapping:
    """Split a string into key-value pairs."""
    try:
        texts = split_str(text, separator=list_separator, brackets=brackets)
    except SplitStrError as error:
        raise _SplitKeyValuePairsSplitError(text=text, inner=error.text) from None
    try:
        pairs = [
            split_str(text_i, separator=pair_separator, brackets=brackets, n=2)
            for text_i in texts
        ]
    except SplitStrError as error:
        raise _SplitKeyValuePairsSplitError(text=text, inner=error.text) from None
    if not mapping:
        return pairs
    try:
        check_duplicates(k for k, _ in pairs)
    except CheckDuplicatesError as error:
        raise _SplitKeyValuePairsDuplicateKeysError(
            text=text, counts=error.counts
        ) from None
    return dict(pairs)


@dataclass(kw_only=True, slots=True)
class SplitKeyValuePairsError(Exception):
    text: str


@dataclass(kw_only=True, slots=True)
class _SplitKeyValuePairsSplitError(SplitKeyValuePairsError):
    inner: str

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r} into key-value pairs"


@dataclass(kw_only=True, slots=True)
class _SplitKeyValuePairsDuplicateKeysError(SplitKeyValuePairsError):
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r} into a mapping since there are duplicate keys; got {get_repr(self.counts)}"


##


@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[1],
) -> tuple[str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[2],
) -> tuple[str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[3],
) -> tuple[str, str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[4],
) -> tuple[str, str, str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: Literal[5],
) -> tuple[str, str, str, str, str]: ...
@overload
def split_str(
    text: str,
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: int | None = None,
) -> tuple[str, ...]: ...
def split_str(
    text: str,
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = None,
    n: int | None = None,
) -> tuple[str, ...]:
    """Split a string, with a special provision for the empty string."""
    if text == "":
        texts = []
    elif text == _escape_separator(separator=separator):
        texts = [""]
    elif brackets is None:
        texts = text.split(separator)
    else:
        texts = _split_str_brackets(text, brackets, separator=separator)
    if n is None:
        return tuple(texts)
    if len(texts) != n:
        raise _SplitStrCountError(text=text, n=n, texts=texts)
    return tuple(texts)


def _split_str_brackets(
    text: str,
    brackets: Iterable[tuple[str, str]],
    /,
    *,
    separator: str = _DEFAULT_SEPARATOR,
) -> list[str]:
    brackets = list(brackets)
    opens, closes = transpose(brackets)
    close_to_open = {close: open_ for open_, close in brackets}

    escapes = map(escape, chain(chain.from_iterable(brackets), [separator]))
    pattern = re.compile("|".join(escapes))

    results: list[str] = []
    stack: deque[tuple[str, int]] = deque()
    last = 0

    for match in pattern.finditer(text):
        token, position = match.group(), match.start()
        if token in opens:
            stack.append((token, position))
        elif token in closes:
            if len(stack) == 0:
                raise _SplitStrClosingBracketUnmatchedError(
                    text=text, token=token, position=position
                )
            open_token, open_position = stack.pop()
            if open_token != close_to_open[token]:
                raise _SplitStrClosingBracketMismatchedError(
                    text=text,
                    opening_token=open_token,
                    opening_position=open_position,
                    closing_token=token,
                    closing_position=position,
                )
        elif (token == separator) and (len(stack) == 0):
            results.append(text[last:position].strip())
            last = position + 1
    results.append(text[last:].strip())
    if len(stack) >= 1:
        token, position = stack.pop()
        raise _SplitStrOpeningBracketUnmatchedError(
            text=text, token=token, position=position
        )
    return results


@dataclass(kw_only=True, slots=True)
class SplitStrError(Exception):
    text: str


@dataclass(kw_only=True, slots=True)
class _SplitStrCountError(SplitStrError):
    n: int
    texts: list[str]

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r} into {self.n} part(s); got {len(self.texts)}"


@dataclass(kw_only=True, slots=True)
class _SplitStrClosingBracketMismatchedError(SplitStrError):
    opening_token: str
    opening_position: int
    closing_token: str
    closing_position: int

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r}; got mismatched {self.opening_token!r} at position {self.opening_position} and {self.closing_token!r} at position {self.closing_position}"


@dataclass(kw_only=True, slots=True)
class _SplitStrClosingBracketUnmatchedError(SplitStrError):
    token: str
    position: int

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r}; got unmatched {self.token!r} at position {self.position}"


@dataclass(kw_only=True, slots=True)
class _SplitStrOpeningBracketUnmatchedError(SplitStrError):
    token: str
    position: int

    @override
    def __str__(self) -> str:
        return f"Unable to split {self.text!r}; got unmatched {self.token!r} at position {self.position}"


def join_strs(
    texts: Iterable[str], /, *, sort: bool = False, separator: str = _DEFAULT_SEPARATOR
) -> str:
    """Join a collection of strings, with a special provision for the empty list."""
    texts = list(texts)
    if sort:
        texts = sorted(texts)
    if texts == []:
        return ""
    if texts == [""]:
        return _escape_separator(separator=separator)
    return separator.join(texts)


def _escape_separator(*, separator: str = _DEFAULT_SEPARATOR) -> str:
    return f"\\{separator}"


##


class secret_str(str):  # noqa: N801
    """A string with an obfuscated representation."""

    __slots__ = ("_text",)
    _REPR: ClassVar[str] = "***"

    def __init__(self, text: str, /) -> None:
        super().__init__()
        self._text = text

    @override
    def __repr__(self) -> str:
        return self._REPR

    @override
    def __str__(self) -> str:
        return self._REPR

    @property
    def str(self) -> str:
        return self._text


##


def str_encode(obj: Any, /) -> bytes:
    """Return the string representation of the object encoded as bytes."""
    return str(obj).encode()


##


def strip_and_dedent(text: str, /, *, trailing: bool = False) -> str:
    """Strip and dedent a string."""
    result = dedent(text.strip("\n")).strip("\n")
    return f"{result}\n" if trailing else result


##


@overload
def to_bool(bool_: MaybeCallableBoolLike, /) -> bool: ...
@overload
def to_bool(bool_: None, /) -> None: ...
@overload
def to_bool(bool_: Sentinel, /) -> Sentinel: ...
def to_bool(
    bool_: MaybeCallableBoolLike | None | Sentinel, /
) -> bool | None | Sentinel:
    """Convert to a bool."""
    match bool_:
        case bool() | None | Sentinel():
            return bool_
        case str():
            return parse_bool(bool_)
        case Callable() as func:
            return to_bool(func())
        case never:
            assert_never(never)


##


@overload
def to_str(text: MaybeCallableStr, /) -> str: ...
@overload
def to_str(text: None, /) -> None: ...
@overload
def to_str(text: Sentinel, /) -> Sentinel: ...
def to_str(text: MaybeCallableStr | None | Sentinel, /) -> str | None | Sentinel:
    """Convert to a string."""
    match text:
        case str() | None | Sentinel():
            return text
        case Callable() as func:
            return to_str(func())
        case never:
            assert_never(never)


##


def unique_str() -> str:
    """Generate at unique string."""
    now = time_ns()
    pid = getpid()
    ident = get_ident()
    key = str(uuid4()).replace("-", "")
    return f"{now}_{pid}_{ident}_{key}"


##


def _kebab_snake_case(text: str, separator: str, /) -> str:
    """Convert text into kebab/snake case."""
    leading = bool(search(r"^_", text))
    trailing = bool(search(r"_$", text))
    parts = _SPLIT_TEXT.findall(text)
    parts = (p for p in parts if len(p) >= 1)
    parts = chain([""] if leading else [], parts, [""] if trailing else [])
    return separator.join(parts).lower()


_SPLIT_TEXT = re.compile(
    r"""
    [A-Z]+(?=[A-Z][a-z0-9]) | # all caps followed by Upper+lower or digit (API in APIResponse2)
    [A-Z]?[a-z]+[0-9]*      | # normal words with optional trailing digits (Text123)
    [A-Z]+[0-9]*            | # consecutive caps with optional trailing digits (ID2)
    """,
    flags=VERBOSE,
)

__all__ = [
    "BRACKETS",
    "LIST_SEPARATOR",
    "PAIR_SEPARATOR",
    "ParseBoolError",
    "ParseNoneError",
    "SplitKeyValuePairsError",
    "SplitStrError",
    "join_strs",
    "kebab_case",
    "parse_bool",
    "parse_none",
    "pascal_case",
    "prompt_bool",
    "repr_encode",
    "secret_str",
    "snake_case",
    "split_key_value_pairs",
    "split_str",
    "str_encode",
    "strip_and_dedent",
    "to_bool",
    "to_str",
    "unique_str",
]



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, cast

from tzlocal import get_localzone

if TYPE_CHECKING:
    from zoneinfo import ZoneInfo

    from utilities.types import TimeZone


def get_local_time_zone() -> ZoneInfo:
    """Get the local time zone, with the logging disabled."""
    logger = getLogger("tzlocal")  # avoid import cycle
    init_disabled = logger.disabled
    logger.disabled = True
    time_zone = get_localzone()
    logger.disabled = init_disabled
    return time_zone


LOCAL_TIME_ZONE = get_local_time_zone()
LOCAL_TIME_ZONE_NAME = cast("TimeZone", LOCAL_TIME_ZONE.key)


__all__ = ["LOCAL_TIME_ZONE", "LOCAL_TIME_ZONE_NAME", "get_local_time_zone"]



from __future__ import annotations

from functools import reduce
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, assert_never, cast, override

from pydantic import Field, create_model
from pydantic_settings import (
    BaseSettings,
    CliSettingsSource,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
    YamlConfigSettingsSource,
)
from pydantic_settings.sources import DEFAULT_PATH

from utilities.errors import ImpossibleCaseError
from utilities.iterables import always_iterable

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from pydantic_settings.sources import PathType

    from utilities.types import MaybeSequenceStr, PathLike


type PathLikeWithSection = tuple[PathLike, MaybeSequenceStr]
type PathLikeOrWithSection = PathLike | PathLikeWithSection


class CustomBaseSettings(BaseSettings):
    """Base settings for loading JSON/TOML/YAML files."""

    # paths
    json_files: ClassVar[Sequence[PathLikeOrWithSection]] = []
    toml_files: ClassVar[Sequence[PathLikeOrWithSection]] = []
    yaml_files: ClassVar[Sequence[PathLikeOrWithSection]] = []

    # config
    model_config: ClassVar[SettingsConfigDict] = SettingsConfigDict(
        frozen=True, env_nested_delimiter="__"
    )

    @classmethod
    @override
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        _ = (init_settings, dotenv_settings, file_secret_settings)
        return tuple(cls._yield_base_settings_sources(settings_cls, env_settings))

    @classmethod
    def _yield_base_settings_sources(
        cls,
        settings_cls: type[BaseSettings],
        env_settings: PydanticBaseSettingsSource,
        /,
    ) -> Iterator[PydanticBaseSettingsSource]:
        yield env_settings
        for file, section in map(_ensure_section, cls.json_files):
            yield JsonConfigSectionSettingsSource(
                settings_cls, json_file=file, section=section
            )
        for file, section in map(_ensure_section, cls.toml_files):
            yield TomlConfigSectionSettingsSource(
                settings_cls, toml_file=file, section=section
            )
        for file, section in map(_ensure_section, cls.yaml_files):
            yield YamlConfigSectionSettingsSource(
                settings_cls, yaml_file=file, section=section
            )


class JsonConfigSectionSettingsSource(JsonConfigSettingsSource):
    @override
    def __init__(
        self,
        settings_cls: type[BaseSettings],
        json_file: PathType | None = DEFAULT_PATH,
        json_file_encoding: str | None = None,
        *,
        section: MaybeSequenceStr,
    ) -> None:
        super().__init__(
            settings_cls, json_file=json_file, json_file_encoding=json_file_encoding
        )
        self.section = section

    @override
    def __call__(self) -> dict[str, Any]:
        return _get_section(super().__call__(), self.section)


class TomlConfigSectionSettingsSource(TomlConfigSettingsSource):
    @override
    def __init__(
        self,
        settings_cls: type[BaseSettings],
        toml_file: PathType | None = DEFAULT_PATH,
        *,
        section: MaybeSequenceStr,
    ) -> None:
        super().__init__(settings_cls, toml_file=toml_file)
        self.section = section

    @override
    def __call__(self) -> dict[str, Any]:
        return _get_section(super().__call__(), self.section)


class YamlConfigSectionSettingsSource(YamlConfigSettingsSource):
    @override
    def __init__(
        self,
        settings_cls: type[BaseSettings],
        yaml_file: PathType | None = DEFAULT_PATH,
        yaml_file_encoding: str | None = None,
        yaml_config_section: str | None = None,
        *,
        section: MaybeSequenceStr,
    ) -> None:
        super().__init__(
            settings_cls,
            yaml_file=yaml_file,
            yaml_file_encoding=yaml_file_encoding,
            yaml_config_section=yaml_config_section,
        )
        self.section = section

    @override
    def __call__(self) -> dict[str, Any]:
        return _get_section(super().__call__(), self.section)


def _ensure_section(file: PathLikeOrWithSection, /) -> PathLikeWithSection:
    match file:
        case Path() | str():
            return file, []
        case Path() | str() as path, str() | list() | tuple() as section:
            return path, section
        case never:
            assert_never(never)


def _get_section(
    mapping: dict[str, Any], section: MaybeSequenceStr, /
) -> dict[str, Any]:
    return reduce(lambda acc, el: acc.get(el, {}), always_iterable(section), mapping)


##


class HashableBaseSettings(BaseSettings):
    """Base settings for loading JSON files."""

    # config
    model_config: ClassVar[SettingsConfigDict] = SettingsConfigDict(frozen=True)


##


def load_settings[T: BaseSettings](cls: type[T], /, *, cli: bool = False) -> T:
    """Load a set of settings."""
    _ = cls.model_rebuild()
    if cli:
        cls_with_defaults = _load_settings_create_model(cls)

        @classmethod
        def settings_customise_sources(
            cls: type[BaseSettings],
            settings_cls: type[BaseSettings],
            init_settings: PydanticBaseSettingsSource,
            env_settings: PydanticBaseSettingsSource,
            dotenv_settings: PydanticBaseSettingsSource,
            file_secret_settings: PydanticBaseSettingsSource,
        ) -> tuple[PydanticBaseSettingsSource, ...]:
            parent = cast(
                "Any", super(cls_with_defaults, cls)
            ).settings_customise_sources(
                settings_cls=settings_cls,
                init_settings=init_settings,
                env_settings=env_settings,
                dotenv_settings=dotenv_settings,
                file_secret_settings=file_secret_settings,
            )
            return (
                CliSettingsSource(
                    settings_cls, cli_parse_args=True, case_sensitive=False
                ),
                *parent,
            )

        cls_use = type(
            cls.__name__,
            (cls_with_defaults,),
            {"settings_customise_sources": settings_customise_sources},
        )
        cls_use = cast("type[T]", cls_use)
    else:
        cls_use = cls
    return cls_use()


def _load_settings_create_model[T: BaseSettings](
    cls: type[T], /, *, values: T | None = None
) -> type[T]:
    values_use = cls() if values is None else values
    kwargs: dict[str, Any] = {}
    for name, field in cls.model_fields.items():
        if (ann := field.annotation) is None:
            raise ImpossibleCaseError(case=[f"{ann=}"])  # pragma: no cover
        value = getattr(values_use, name)
        if (
            isinstance(cast("Any", ann), type)  # 'ann' is possible not a type
            and issubclass(ann, BaseSettings)
        ):
            kwargs[name] = _load_settings_create_model(ann, values=value)
        else:
            kwargs[name] = (field.annotation, Field(default=value))
    return create_model(cls.__name__, __base__=cls, **kwargs)


__all__ = [
    "CustomBaseSettings",
    "HashableBaseSettings",
    "JsonConfigSectionSettingsSource",
    "TomlConfigSectionSettingsSource",
    "YamlConfigSectionSettingsSource",
    "load_settings",
]



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass
from typing import TYPE_CHECKING, assert_never, cast, override
from zoneinfo import ZoneInfo

from whenever import ZonedDateTime

from utilities.types import TIME_ZONES
from utilities.tzlocal import LOCAL_TIME_ZONE, LOCAL_TIME_ZONE_NAME

if TYPE_CHECKING:
    from utilities.types import TimeZone, TimeZoneLike


UTC = ZoneInfo("UTC")


##


def to_zone_info(obj: TimeZoneLike, /) -> ZoneInfo:
    """Convert to a time-zone."""
    match obj:
        case ZoneInfo() as zone_info:
            return zone_info
        case ZonedDateTime() as date_time:
            return ZoneInfo(date_time.tz)
        case "local" | "localtime":
            return LOCAL_TIME_ZONE
        case str() as key:
            return ZoneInfo(key)
        case dt.tzinfo() as tzinfo:
            if tzinfo is dt.UTC:
                return UTC
            raise _ToZoneInfoInvalidTZInfoError(time_zone=obj)
        case dt.datetime() as date_time:
            if date_time.tzinfo is None:
                raise _ToZoneInfoPlainDateTimeError(date_time=date_time)
            return to_zone_info(date_time.tzinfo)
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToTimeZoneError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToZoneInfoInvalidTZInfoError(ToTimeZoneError):
    time_zone: dt.tzinfo

    @override
    def __str__(self) -> str:
        return f"Invalid time-zone: {self.time_zone}"


@dataclass(kw_only=True, slots=True)
class _ToZoneInfoPlainDateTimeError(ToTimeZoneError):
    date_time: dt.datetime

    @override
    def __str__(self) -> str:
        return f"Plain date-time: {self.date_time}"


##


def to_time_zone_name(obj: TimeZoneLike, /) -> TimeZone:
    """Convert to a time zone name."""
    match obj:
        case ZoneInfo() as zone_info:
            return cast("TimeZone", zone_info.key)
        case ZonedDateTime() as date_time:
            return cast("TimeZone", date_time.tz)
        case "local" | "localtime":
            return LOCAL_TIME_ZONE_NAME
        case str() as time_zone:
            if time_zone in TIME_ZONES:
                return time_zone
            raise _ToTimeZoneNameInvalidKeyError(time_zone=time_zone)
        case dt.tzinfo() as tzinfo:
            if tzinfo is dt.UTC:
                return cast("TimeZone", UTC.key)
            raise _ToTimeZoneNameInvalidTZInfoError(time_zone=obj)
        case dt.datetime() as date_time:
            if date_time.tzinfo is None:
                raise _ToTimeZoneNamePlainDateTimeError(date_time=date_time)
            return to_time_zone_name(date_time.tzinfo)
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ToTimeZoneNameError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ToTimeZoneNameInvalidKeyError(ToTimeZoneNameError):
    time_zone: str

    @override
    def __str__(self) -> str:
        return f"Invalid time-zone: {self.time_zone!r}"


@dataclass(kw_only=True, slots=True)
class _ToTimeZoneNameInvalidTZInfoError(ToTimeZoneNameError):
    time_zone: dt.tzinfo

    @override
    def __str__(self) -> str:
        return f"Invalid time-zone: {self.time_zone}"


@dataclass(kw_only=True, slots=True)
class _ToTimeZoneNamePlainDateTimeError(ToTimeZoneNameError):
    date_time: dt.datetime

    @override
    def __str__(self) -> str:
        return f"Plain date-time: {self.date_time}"


__all__ = [
    "UTC",
    "ToTimeZoneError",
    "ToTimeZoneNameError",
    "to_time_zone_name",
    "to_zone_info",
]



from __future__ import annotations

from contextlib import suppress
from json import loads
from pathlib import Path
from shutil import copytree
from typing import TYPE_CHECKING, Any, assert_never

from pytest_regressions.file_regression import FileRegressionFixture

from utilities.functions import ensure_str
from utilities.operator import is_equal

if TYPE_CHECKING:
    from polars import DataFrame, Series
    from pytest import FixtureRequest

    from utilities.types import PathLike, StrMapping


##


class OrjsonRegressionFixture:
    """Implementation of `orjson_regression` fixture."""

    def __init__(
        self, path: PathLike, request: FixtureRequest, tmp_path: Path, /
    ) -> None:
        super().__init__()
        path = Path(path)
        original_datadir = path.parent
        data_dir = tmp_path.joinpath(ensure_str(request.fixturename))
        with suppress(FileNotFoundError):
            _ = copytree(original_datadir, data_dir)
        self._fixture = FileRegressionFixture(
            datadir=data_dir, original_datadir=original_datadir, request=request
        )
        self._basename = path.name

    def check(
        self,
        obj: Any,
        /,
        *,
        globalns: StrMapping | None = None,
        localns: StrMapping | None = None,
        warn_name_errors: bool = False,
        dataclass_defaults: bool = False,
        suffix: str | None = None,
    ) -> None:
        """Check the serialization of the object against the baseline."""
        from utilities.orjson import serialize

        data = serialize(
            obj,
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
            dataclass_defaults=dataclass_defaults,
        )
        basename = self._basename
        if suffix is not None:
            basename = f"{basename}__{suffix}"
        self._fixture.check(
            data,
            extension=".json",
            basename=basename,
            binary=True,
            check_fn=self._check_fn,
        )

    def _check_fn(self, path1: Path, path2: Path, /) -> None:
        left = loads(path1.read_text())
        right = loads(path2.read_text())
        assert is_equal(left, right), f"{left=}, {right=}"


##


class PolarsRegressionFixture:
    """Implementation of `polars_regression`."""

    def __init__(
        self, path: PathLike, request: FixtureRequest, tmp_path: Path, /
    ) -> None:
        super().__init__()
        self._fixture = OrjsonRegressionFixture(path, request, tmp_path)

    def check(self, obj: Series | DataFrame, /, *, suffix: str | None = None) -> None:
        """Check the Series/DataFrame summary against the baseline."""
        from polars import DataFrame, Series, col
        from polars.exceptions import InvalidOperationError

        data: StrMapping = {
            "describe": obj.describe(percentiles=[i / 10 for i in range(1, 10)]).rows(
                named=True
            ),
            "is_empty": obj.is_empty(),
            "n_unique": obj.n_unique(),
        }
        match obj:
            case Series() as series:
                data["has_nulls"] = series.has_nulls()
                data["is_sorted"] = series.is_sorted()
                data["len"] = series.len()
                data["null_count"] = series.null_count()
            case DataFrame() as df:
                approx_n_unique: dict[str, int] = {}
                for column in df.columns:
                    with suppress(InvalidOperationError):
                        approx_n_unique[column] = df.select(
                            col(column).approx_n_unique()
                        ).item()
                data["approx_n_unique"] = approx_n_unique
                data["glimpse"] = df.glimpse(return_type="string")
                data["null_count"] = df.null_count().row(0, named=True)
            case never:
                assert_never(never)
        self._fixture.check(data, suffix=suffix)


__all__ = ["OrjsonRegressionFixture", "PolarsRegressionFixture"]



from __future__ import annotations

import sys
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from string import Template
from subprocess import PIPE, CalledProcessError, Popen
from threading import Thread
from typing import IO, TYPE_CHECKING, Literal, assert_never, overload

from utilities.errors import ImpossibleCaseError

if TYPE_CHECKING:
    from collections.abc import Iterator

    from utilities.types import PathLike, StrMapping, StrStrMapping


MKTEMP_DIR_CMD = ["mktemp", "-d"]


def echo_cmd(text: str, /) -> list[str]:
    return ["echo", text]


def expand_path(
    path: PathLike, /, *, subs: StrMapping | None = None, sudo: bool = False
) -> Path:
    if subs is not None:
        path = Template(str(path)).substitute(**subs)
    if sudo:  # pragma: no cover
        return Path(run(*sudo_cmd(*echo_cmd(str(path))), return_=True))
    return Path(path).expanduser()


def maybe_sudo_cmd(cmd: str, /, *args: str, sudo: bool = False) -> list[str]:
    parts: list[str] = [cmd, *args]
    return sudo_cmd(*parts) if sudo else parts


def mkdir(path: PathLike, /, *, sudo: bool = False, parent: bool = False) -> None:
    if sudo:  # pragma: no cover
        run(*sudo_cmd(*mkdir_cmd(path, parent=parent)))
    else:
        path = expand_path(path)
        path_use = path.parent if parent else path
        path_use.mkdir(parents=True, exist_ok=True)


def mkdir_cmd(path: PathLike, /, *, parent: bool = False) -> list[str]:
    path_use = f"$(dirname {path})" if parent else path
    return ["mkdir", "-p", str(path_use)]


def rm_cmd(path: PathLike, /) -> list[str]:
    return ["rm", "-rf", str(path)]


@overload
def run(
    cmd: str,
    /,
    *cmds: str,
    executable: str | None = None,
    shell: bool = False,
    cwd: PathLike | None = None,
    env: StrStrMapping | None = None,
    user: str | int | None = None,
    group: str | int | None = None,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: Literal[True],
    return_stdout: bool = False,
    return_stderr: bool = False,
) -> str: ...
@overload
def run(
    cmd: str,
    /,
    *cmds: str,
    executable: str | None = None,
    shell: bool = False,
    cwd: PathLike | None = None,
    env: StrStrMapping | None = None,
    user: str | int | None = None,
    group: str | int | None = None,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: Literal[True],
    return_stderr: bool = False,
) -> str: ...
@overload
def run(
    cmd: str,
    /,
    *cmds: str,
    executable: str | None = None,
    shell: bool = False,
    cwd: PathLike | None = None,
    env: StrStrMapping | None = None,
    user: str | int | None = None,
    group: str | int | None = None,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: bool = False,
    return_stderr: Literal[True],
) -> str: ...
@overload
def run(
    cmd: str,
    /,
    *cmds: str,
    executable: str | None = None,
    shell: bool = False,
    cwd: PathLike | None = None,
    env: StrStrMapping | None = None,
    user: str | int | None = None,
    group: str | int | None = None,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: Literal[False] = False,
    return_stdout: Literal[False] = False,
    return_stderr: Literal[False] = False,
) -> None: ...
@overload
def run(
    cmd: str,
    /,
    *cmds: str,
    executable: str | None = None,
    shell: bool = False,
    cwd: PathLike | None = None,
    env: StrStrMapping | None = None,
    user: str | int | None = None,
    group: str | int | None = None,
    print: bool = False,
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: bool = False,
    return_stderr: bool = False,
) -> str | None: ...
def run(
    cmd: str,
    /,
    *cmds: str,
    executable: str | None = None,
    shell: bool = False,
    cwd: PathLike | None = None,
    env: StrStrMapping | None = None,
    user: str | int | None = None,
    group: str | int | None = None,
    print: bool = False,  # noqa: A002
    print_stdout: bool = False,
    print_stderr: bool = False,
    return_: bool = False,
    return_stdout: bool = False,
    return_stderr: bool = False,
) -> str | None:
    buffer = StringIO()
    stdout = StringIO()
    stderr = StringIO()
    with Popen(
        [cmd, *cmds],
        bufsize=1,
        executable=executable,
        stdout=PIPE,
        stderr=PIPE,
        shell=shell,
        cwd=cwd,
        env=env,
        text=True,
        user=user,
        group=group,
    ) as proc:
        if proc.stdout is None:  # pragma: no cover
            raise ImpossibleCaseError(case=[f"{proc.stdout=}"])
        if proc.stderr is None:  # pragma: no cover
            raise ImpossibleCaseError(case=[f"{proc.stderr=}"])
        with (
            _yield_write(
                proc.stdout,
                buffer,
                stdout,
                *([sys.stdout] if print or print_stdout else []),
            ),
            _yield_write(
                proc.stderr,
                buffer,
                stderr,
                *([sys.stderr] if print or print_stderr else []),
            ),
        ):
            return_code = proc.wait()
        match return_code, return_ or return_stdout, return_ or return_stderr:
            case 0, True, True:
                _ = buffer.seek(0)
                return buffer.read()
            case 0, True, False:
                _ = stdout.seek(0)
                return stdout.read()
            case 0, False, True:
                _ = stderr.seek(0)
                return stderr.read()
            case 0, False, False:
                return None
            case _, _, _:
                _ = stdout.seek(0)
                _ = stderr.seek(0)
                raise CalledProcessError(
                    return_code, cmd, output=stdout.read(), stderr=stderr.read()
                )
            case never:
                assert_never(never)


@contextmanager
def _yield_write(input_: IO[str], /, *outputs: IO[str]) -> Iterator[None]:
    thread = Thread(target=_run_target, args=(input_, *outputs), daemon=True)
    thread.start()
    try:
        yield
    finally:
        thread.join()


def _run_target(input_: IO[str], /, *outputs: IO[str]) -> None:
    with input_:
        for line in iter(input_.readline, ""):
            for output in outputs:
                _ = output.write(line)


def sudo_cmd(cmd: str, /, *args: str) -> list[str]:
    return ["sudo", cmd, *args]


def touch_cmd(path: PathLike, /) -> list[str]:
    return ["touch", str(path)]


__all__ = [
    "MKTEMP_DIR_CMD",
    "echo_cmd",
    "maybe_sudo_cmd",
    "mkdir",
    "mkdir_cmd",
    "rm_cmd",
    "run",
    "sudo_cmd",
    "touch_cmd",
]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from testbook import testbook

from utilities.pytest import throttle
from utilities.text import pascal_case

if TYPE_CHECKING:
    from collections.abc import Callable

    from utilities.types import Delta, PathLike


def build_notebook_tester(
    path: PathLike, /, *, throttle: Delta | None = None, on_try: bool = False
) -> type[Any]:
    """Build the notebook tester class."""
    path = Path(path)
    name = f"Test{pascal_case(path.stem)}"
    notebooks = [
        path_i
        for path_i in path.rglob("**/*.ipynb")
        if all(p != ".ipynb_checkpoints" for p in path_i.parts)
    ]
    namespace = {
        f"test_{p.stem.replace('-', '_')}": _build_test_method(
            p, delta=throttle, on_try=on_try
        )
        for p in notebooks
    }
    return type(name, (), namespace)


def _build_test_method(
    path: Path, /, *, delta: Delta | None = None, on_try: bool = False
) -> Callable[..., Any]:
    @testbook(path, execute=True)
    def method(self: Any, tb: Any) -> None:
        _ = (self, tb)  # pragma: no cover

    if delta is not None:
        method = throttle(delta=delta, on_try=on_try)(method)

    return method


__all__ = ["build_notebook_tester"]
